################################################
#
#  $Id: Container.pm,v 1.5 2003/01/19 02:39:57 joko Exp $
#
#  $Log: Container.pm,v $
#  Revision 1.5  2003/01/19 02:39:57  joko
#  + moved 'deep_copy' from module 'libp' to module 'Data::Transform::Deep'
#  + preserved order for hashes '$self->{config}' and '$self->{locator}' by using Tie::IxHash
#
#  Revision 1.4  2002/12/04 07:38:07  jonen
#  + deep copy
#
#  Revision 1.3  2002/12/01 22:18:28  joko
#  - no interactive implicit deploy
#  + only test integrity if storage available
#
#  Revision 1.2  2002/12/01 07:08:35  joko
#  + needs to "use Data::Storage;"
#
#  Revision 1.1  2002/11/29 04:54:54  joko
#  + initial check-in
#
#  Revision 1.1  2002/11/17 06:20:41  joko
#  + initial check in
#
#
################################################


package Data::Storage::Container;

use strict;
use warnings;

# get logger instance
my $logger = Log::Dispatch::Config->instance;


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -   main
use Tie::IxHash;
use Data::Dumper;

use Data::Transform::Deep qw( deep_copy );
use Data::Storage;
#use Data::Storage::Locator;

sub new {
  my $invocant = shift;
  my $class = ref($invocant) || $invocant;

  # get constructor arguments
  my @args = ();
  @_ && (@args = @_);
  $logger->debug( __PACKAGE__ . "->new( @args )" );

  my $self = { @_ };
  bless $self, $class;
  
  # preserve order of configuration variables
  #$self->{config} = {};
  tie %{$self->{config}}, 'Tie::IxHash';
  tie %{$self->{locator}}, 'Tie::IxHash';
  
  return $self;
}

sub addConfig {
  my $self = shift;
  my $name = shift;
  my $config = shift;
  $self->{config}->{$name} = $config;
}

sub addConfigByDsn {
  my $self = shift;
  my $name = shift;
  my $dsn = shift;
  
  # HACK: assume DBI as default type for now
  # TODO: guess type from dsn
  my $type = "DBI";
  
  $self->addConfig($name, { type => $type, dsn => $dsn });
}

sub addStorage {
  my $self = shift;
  my $name = shift;
  my $storage = shift;
  $self->{storage}->{$name} = $storage;
}

sub initLocator {
  my $self = shift;
  my $name = shift;
  my $db_config = $self->{config}->{$name};

  $logger->debug( __PACKAGE__ . "->initLocator( db_config $db_config )" );

  my $cfg_locator = {};
  
  # set default settings, if any
  my $default = deep_copy($self->{config}->{_default});
  foreach (keys %$default) {
    $cfg_locator->{$_} = $default->{$_};
  }
  
  # name it
  $cfg_locator->{name} = $name;
  
  # merge in specific settings
  my $specific = deep_copy($db_config);
  foreach (keys %$specific) {
    $cfg_locator->{$_} = $specific->{$_};
  }
  
  # HACK: transfer dsn from main to dbi settings
  $cfg_locator->{dbi}->{dsn} = $cfg_locator->{dsn} if $cfg_locator->{dsn};

  # HACK: set classnames empty if none given from config
  $cfg_locator->{classnames} = [] if !$cfg_locator->{classnames};

  # HACK: set errorhandler if dbi settings are present
  $cfg_locator->{dbi}->{HandleError} = \&_dbErrorHandler if $cfg_locator->{dbi};

  # create new locator object
  $self->{locator}->{$name} = Data::Storage::Locator->new( $cfg_locator );

}

sub initLocators {
  my $self = shift;
  foreach (keys %{$self->{config}}) {
    #print $_, "\n";
    $self->initLocator($_, $self->{config}->{$_}) if !/^_/;
  }
  #print "locs: ", Dumper($self->{locator});
}

sub getLocator {
  my $self = shift;
  my $name = shift;
  return $self->{locator}->{$name};
}

sub initStorage {
  my $self = shift;
  my $name = shift;

  my $locator = $self->getLocator($name);
  $logger->info( __PACKAGE__ . " is booting storage declared by locator \"$name\"" );

  my $storage = Data::Storage->new($locator);

  my $log_prefix = __PACKAGE__ . "->initStorage: ";
  $log_prefix .= "dsn=\"$self->{locator}->{$name}->{dsn}\"" if $self->{locator}->{$name}->{dsn};

  # should we test availability of the storage before using it?
  if ($locator->{test_availability}) {
    $locator->{status}->{availability} = $storage->testAvailability();
    if ( !$storage->testAvailability() ) {
      $logger->error( "$log_prefix: testAvailability failed" );
    }
  }
          
  # should we test integrity of the storage before using it?
  if ($locator->{status}->{availability} && $locator->{test_integrity}) {
    #return unless $storage->testIntegrity();
    $locator->{status}->{integrity} = $storage->testIntegrity();
    # actions if integrity fails
    if (!$locator->{status}->{integrity}) {
      $logger->error( "$log_prefix: testIntegrity failed" );
    }
  }

  # try to connect...
  # TODO:
  # don't connect right here, do an implicit connect on (later) usage
  # maybe set ->{meta}->{connectmethod} = "connect" here
  #return unless $storage->connect();
  $storage->connect() if $locator->{status}->{integrity};

  # should we check emptyness?
  if ($locator->{status}->{availability} && $locator->{test_emptyness}) {
    #print "test empty", "\n";
    if ( !@{$storage->getChildNodes()} ) {
      $locator->{status}->{empty} = 1;
      $logger->warning( "$log_prefix: Storage is empty.");
      #return;
    }
  }

  # expand logging?
  if ( $locator->{status}->{integrity} && $locator->{logger} ) {
    # expand logging (to Tangram-Database)
    # TODO:
    # - move configuration data from this code to db_config somehow
    # - do complete handling of this stuff in Data::Storage::* also
    # - just leave a simple on/off-trigger here and/or 
    # - make it optionally configurable via e.g. get/set-methods
    #    to satisfy "manual"-mode instead of "config"-mode
    # - (re-)add hierarchical logging (each event may have a parent)
    no strict;
    my $creator = sub { return new SystemEvent; };
    use strict;
    $storage->addLogDispatchHandler("Tangram11", "Log::Dispatch::Tangram", $logger, $creator);
    #$storage_left->addLogDispatchHandler("Tangram11", "Log::Dispatch::Tangram", $logger);
  }

  #$self->{storage}->{$name} = $storage;
  #print "add storage: $name", "\n";
  $self->addStorage($name, $storage);

  return 1;
  
}

sub initStorages {
  my $self = shift;
  foreach (keys %{$self->{locator}}) {
    #print $_, "\n";
    $self->initStorage($_);
  }
}

sub _dbErrorHandler {
  my $message_db = shift;
  # get logger instance
  #my $logger = Log::Dispatch::Config->instance;
  #$logger->log(level => 'error', message => "Tangram DBI-Error: $message");
  my $message_log = "DBI-Error: $message_db";
  print STDERR $message_log, "\n";
  #$logger->error( $message_log );
  #$logger->log_to( name => 'file', level => 'error', message => $message_log);
  $logger->log_to( name => 'file', level => 'warning', message => $message_log);
  return 1;
}

1;
