#################################
#
#  $Id: DBI.pm,v 1.11 2002/12/19 16:31:05 joko Exp $
#
#  $Log: DBI.pm,v $
#  Revision 1.11  2002/12/19 16:31:05  joko
#  + sub dropDb
#  + sub rebuildDb
#
#  Revision 1.10  2002/12/15 02:02:22  joko
#  + fixed logging-message
#
#  Revision 1.9  2002/12/05 07:58:20  joko
#  + now using Tie::SecureHash as a base for the COREHANDLE
#  + former public COREHANDLE becomes private _COREHANDLE now
#
#  Revision 1.8  2002/12/01 22:20:43  joko
#  + sub createDb (from Storage.pm)
#
#  Revision 1.7  2002/12/01 07:09:09  joko
#  + sub getListFiltered (dummy redirecting to getListUnfiltered)
#
#  Revision 1.6  2002/12/01 04:46:01  joko
#  + sub eraseAll
#
#  Revision 1.5  2002/11/29 05:00:26  joko
#  + sub getListUnfiltered
#  + sub sendQuery
#
#  Revision 1.4  2002/11/17 08:46:42  jonen
#  + wrapped eval around DBI->connect to prevent deaths
#
#  Revision 1.3  2002/11/17 06:34:39  joko
#  + locator metadata can now be reached via ->{locator}
#  - sub hash2sql now taken from libdb
#
#  Revision 1.2  2002/10/25 11:43:27  joko
#  + enhanced robustness
#  + more logging for debug-levels
#
#  Revision 1.1  2002/10/10 03:44:07  cvsjoko
#  + new
#
#
#################################

package Data::Storage::Handler::DBI;

use strict;
use warnings;

use base ("Data::Storage::Handler::Abstract");

use DBI;
use Data::Dumper;
use libdb qw( getDbNameByDsn hash2Sql );
use Data::Storage::Result::DBI;

# get logger instance
my $logger = Log::Dispatch::Config->instance;


sub getMetaInfo {
  my $self = shift;
  $logger->debug( __PACKAGE__ . "->getMetaInfo()"  );
  return {
    'disconnectMethod' => 'disconnect',
  };
}

sub connect {

    my $self = shift;
    
    # create handle
      if ( my $dsn = $self->{locator}->{dbi}->{dsn} ) {
      #if ( my $dsn = $self->{locator}->{dsn} ) {
        $logger->debug( __PACKAGE__ . "->connect( dsn $dsn )" );
        
        # HACK:
        # set errorhandler before actually calling DBI->connect 
        # in order to catch errors from the very beginning
        #DBI->{HandleError} = $self->{dbi}->{HandleError};
        
        #use Data::Dumper; print Dumper($self->{dbi});
        
        eval {
          $self->{_COREHANDLE} = DBI->connect( $dsn, '', '', $self->{locator}->{dbi} );
          if (!$self->{_COREHANDLE}) {
            $logger->warning( __PACKAGE__ . "->connect failed: " . DBI::errstr );
            return;
          }
        };
        $logger->warning( __PACKAGE__ . "->connect failed: " . $@ ) if $@;

      }
      $self->configureCOREHANDLE();

    $self->{locator}->{status}->{connected} = 1;

    return 1;
    
}

sub configureCOREHANDLE {

  my $self = shift;

  $logger->debug( __PACKAGE__ . "->configureCOREHANDLE" );

  return if !$self->{_COREHANDLE};

  # apply configured modifications to DBI-handle
    if (exists $self->{locator}->{dbi}->{trace_level} && exists $self->{locator}->{dbi}->{trace_file}) {
      $self->{_COREHANDLE}->trace($self->{locator}->{dbi}->{trace_level}, $self->{locator}->{dbi}->{trace_file});
    }
    if (exists $self->{locator}->{dbi}->{RaiseError}) {
      $self->{_COREHANDLE}->{RaiseError} = $self->{locator}->{dbi}->{RaiseError};
    }
    if (exists $self->{locator}->{dbi}->{PrintError}) {
      $self->{_COREHANDLE}->{PrintError} = $self->{locator}->{dbi}->{PrintError};
    }
    if (exists $self->{locator}->{dbi}->{HandleError}) {
      $self->{_COREHANDLE}->{HandleError} = $self->{locator}->{dbi}->{HandleError};
    }

}

sub _sendSql {
  my $self = shift;
  my $sql = shift;
  
  # two-level handling for implicit connect:
  # if there's no corehandle ...
  if (!$self->{_COREHANDLE}) {
    # ... try to connect, but ...
    $self->connect();
    # ... if this still fails, there's something wrong probably, so we won't continue
    if (!$self->{_COREHANDLE}) {
      return;
    }
  }
  
  #print "prepare sql: $sql\n";
  
  my $sth = $self->{_COREHANDLE}->prepare($sql);
  $sth->execute();
  return $sth;
}

sub sendCommand {
  my $self = shift;
  my $command = shift;
  # TODO: when tracing: yes, do actually log this
  #$logger->debug( __PACKAGE__ . "->sendCommand( command $command )" );
  my $cmdHandle = $self->_sendSql($command);
  my $result = Data::Storage::Result::DBI->new( RESULTHANDLE => $cmdHandle );
  return $result;
}

sub getChildNodes {
  my $self = shift;
  my @nodes;
  $logger->debug( __PACKAGE__ . "->getChildNodes()" );
  my $locator = $self->{locator};
  #print Dumper($locator); exit;
  if (my $result = $self->sendCommand( 'SHOW TABLES;' ) ) {
    my $dbname = getDbNameByDsn($self->{locator}->{dbi}->{dsn});
    my $key = "Tables_in_$dbname";
    while ( my $row = $result->getNextEntry() ) {
      push @nodes, $row->{$key};
    }
  }
  return \@nodes;
}

sub getListUnfiltered {
  my $self = shift;
  my $nodename = shift;
  my @list;
  $logger->debug( __PACKAGE__ . "->getListUnfiltered( nodename => '" . $nodename . "' )" );
  # get list of rows from rdbms by table name
  my $result = $self->sendCommand("SELECT * FROM $nodename");
  while ( my $row = $result->getNextEntry() ) {
    push @list, $row;
  }
  return \@list;
}

sub sendQuery {
  my $self = shift;
  my $query = shift;

  $logger->debug( __PACKAGE__ . "->sendQuery" );

  #my $sql = "SELECT cs FROM $self->{metainfo}->{$descent}->{node} WHERE $self->{metainfo}->{$descent}->{IdentProvider}->{arg}='$self->{entry}->{source}->{ident}';";
  #my $result = $self->{metainfo}->{$descent}->{storage}->sendCommand($sql);
  my @crits;
  foreach (@{$query->{criterias}}) {
    my $op = '';
    $op = '=' if lc $_->{op} eq 'eq';
    push @crits, "$_->{key}$op'$_->{val}'";
  }
  my $subnodes = {};
  map { $subnodes->{$_}++ } @{$query->{subnodes}};
  # HACK: this is hardcoded ;(    expand possibilities!
  my $crit = join(' AND ', @crits);
  my $sql = hash2Sql($query->{node}, $subnodes, 'SELECT', $crit);
  return $self->sendCommand($sql);
}

sub eraseAll {
  my $self = shift;
  my $classname = shift;
  $logger->debug( __PACKAGE__ . "->eraseAll" );
  my $sql = "DELETE FROM $classname";
  $self->sendCommand($sql);
}

# TODO: actually implement the filtering functionality using $this->sendQuery
sub getListFiltered {
  my $self = shift;
  my $nodename = shift;
  return $self->getListUnfiltered($nodename);
}

# TODO: do this via a parametrized "$self->connect(<connect just to database server - don't select database>)"
sub createDb {
  
  my $self = shift;
  
  # get dsn from Data::Storage::Locator instance
  my $dsn = $self->{locator}->{dbi}->{dsn};

  $logger->debug( __PACKAGE__ .  "->createDb( dsn $dsn )" );

  # remove database setting from dsn-string
  $dsn =~ s/database=(.+?);//;
  
  # remember extracted database name to know what actually to create right now
  my $database_name = $1;

  # flag to indicate goodness
  my $ok;
  
  # connect to database server - don't select/use any specific database
  #if ( my $dbh = DBI->connect($dsn, '', '', { PrintError => 0 } ) ) {
  if ( my $dbh = DBI->connect($dsn, '', '', $self->{locator}->{dbi} ) ) {

    if ($database_name) {
      if ($dbh->do("CREATE DATABASE $database_name")) {
        $ok = 1;
      }
    }

    $dbh->disconnect();

  }
  
  return $ok;
  
}

sub getCOREHANDLE2 {
  my $self = shift;
  return $self->{_COREHANDLE};
}

sub dropDb {
  my $self = shift;
  my $dsn = $self->{locator}->{dbi}->{dsn};

  $logger->debug( __PACKAGE__ .  "->dropDb( dsn $dsn )" );

  $dsn =~ s/database=(.+?);//;
  my $database_name = $1;

  my $ok;
  
  if ( my $dbh = DBI->connect($dsn, '', '', {
                                                      PrintError => 0,
                                                      } ) ) {
    if ($database_name) {
      if ($dbh->do("DROP DATABASE $database_name;")) {
        $ok = 1;
      }
    }

    $dbh->disconnect();

  }
  
  return $ok;
}

sub rebuildDb {
  my $self = shift;
  $logger->info( __PACKAGE__ . "->rebuildDb()" );
  my @results;

  # sum up results (bool (0/1)) in array
  #push @results, $self->retreatSchema();
  push @results, $self->dropDb();
  push @results, $self->createDb();
  #push @results, $self->deploySchema();

  # scan array for "bad ones"
  my $res = 1;
  map {
    $res = 0 if (!$_);
  } @results;
  
  return $res;
}

1;
